""":mod:`pokerkit.tests.test_state` implements unit tests for
:mod:`pokerkit.state`.
"""

from functools import partial
from hashlib import md5
from itertools import combinations
from unittest import main, TestCase

from pokerkit.games import (
    FixedLimitDeuceToSevenLowballTripleDraw,
    FixedLimitOmahaHoldemHighLowSplitEightOrBetter,
    FixedLimitRazz,
    FixedLimitSevenCardStud,
    NoLimitShortDeckHoldem,
    NoLimitTexasHoldem,
)
from pokerkit.state import (
    Automation,
    _HighHandOpeningLookup,
    _LowHandOpeningLookup,
    State,
)
from pokerkit.tests.test_lookups import LookupTestCaseMixin
from pokerkit.utilities import Deck, rake, ValuesLike


class LowHandOpeningLookupTestCase(LookupTestCaseMixin, TestCase):
    def test_get_entry(self) -> None:
        combinations_ = []

        for i in range(1, 5):
            for combination in combinations(Deck.STANDARD, i):
                combinations_.append(combination)

        lookup = _LowHandOpeningLookup()
        combinations_.sort(key=lookup.get_entry)
        string = self.serialize_combinations(combinations_)
        algorithm = md5()
        algorithm.update(string.encode())

        self.assertEqual(
            algorithm.hexdigest(),
            '9b19d8c42e4b03661329b424d7b6c8e9',
        )


class HighHandOpeningLookupTestCase(LookupTestCaseMixin, TestCase):
    def test_get_entry(self) -> None:
        combinations_ = []

        for i in range(1, 5):
            for combination in combinations(Deck.STANDARD, i):
                combinations_.append(combination)

        lookup = _HighHandOpeningLookup()
        combinations_.sort(key=lookup.get_entry)
        string = self.serialize_combinations(combinations_)
        algorithm = md5()
        algorithm.update(string.encode())

        self.assertEqual(
            algorithm.hexdigest(),
            'b11a0c444a528e78c5be57c7bb6db06c',
        )


class StateTestCase(TestCase):
    def test_all_ins(self) -> None:
        state = NoLimitTexasHoldem.create_state(
            (
                Automation.ANTE_POSTING,
                Automation.BET_COLLECTION,
                Automation.BLIND_OR_STRADDLE_POSTING,
                Automation.HOLE_CARDS_SHOWING_OR_MUCKING,
                Automation.HAND_KILLING,
                Automation.CHIPS_PUSHING,
                Automation.CHIPS_PULLING,
            ),
            False,
            {1: 2},
            (1, 2),
            2,
            (6, 4, 1),
            3,
        )

        state.deal_hole('AcAd')
        state.deal_hole('KcKd')
        state.deal_hole('QcQd')
        self.assertRaises(ValueError, state.complete_bet_or_raise_to)
        state.check_or_call()
        self.assertRaises(ValueError, state.complete_bet_or_raise_to)
        state.check_or_call()
        state.burn_card('??')
        state.deal_board('QhQs2c')
        state.burn_card('??')
        state.deal_board('3c')
        state.burn_card('??')
        state.deal_board('4c')
        self.assertFalse(state.status)
        self.assertEqual(state.stacks, [6, 0, 5])

        state = NoLimitTexasHoldem.create_state(
            (
                Automation.ANTE_POSTING,
                Automation.BET_COLLECTION,
                Automation.BLIND_OR_STRADDLE_POSTING,
                Automation.HOLE_CARDS_SHOWING_OR_MUCKING,
                Automation.HAND_KILLING,
                Automation.CHIPS_PUSHING,
                Automation.CHIPS_PULLING,
            ),
            True,
            2,
            (1, 2),
            2,
            (6, 4, 1),
            3,
        )

        state.deal_hole('AcAd')
        state.deal_hole('KcKd')
        state.deal_hole('QcQd')
        self.assertRaises(ValueError, state.complete_bet_or_raise_to)
        state.check_or_call()
        state.burn_card('??')
        state.deal_board('QhQs2c')
        state.burn_card('??')
        state.deal_board('3c')
        state.burn_card('??')
        state.deal_board('4c')
        self.assertFalse(state.status)
        self.assertEqual(state.stacks, [8, 0, 3])

        state = NoLimitTexasHoldem.create_state(
            (
                Automation.ANTE_POSTING,
                Automation.BET_COLLECTION,
                Automation.BLIND_OR_STRADDLE_POSTING,
                Automation.HOLE_CARDS_SHOWING_OR_MUCKING,
                Automation.HAND_KILLING,
                Automation.CHIPS_PUSHING,
                Automation.CHIPS_PULLING,
            ),
            True,
            2,
            (1, 2),
            2,
            (3, 4, 1),
            3,
        )

        state.deal_hole('AcAd')
        state.deal_hole('KcKd')
        state.deal_hole('QcQd')
        state.burn_card('??')
        state.deal_board('QhQs2c')
        state.burn_card('??')
        state.deal_board('3c')
        state.burn_card('??')
        state.deal_board('4c')
        self.assertFalse(state.status)
        self.assertEqual(state.stacks, [4, 1, 3])

        state = NoLimitTexasHoldem.create_state(
            (
                Automation.ANTE_POSTING,
                Automation.BET_COLLECTION,
                Automation.BLIND_OR_STRADDLE_POSTING,
                Automation.HOLE_CARDS_SHOWING_OR_MUCKING,
                Automation.HAND_KILLING,
                Automation.CHIPS_PUSHING,
                Automation.CHIPS_PULLING,
            ),
            True,
            2,
            (1, 2),
            2,
            (3, 4, 6, 1, 2, 2),
            6,
        )

        state.deal_hole('AcAd')
        state.deal_hole('KcKd')
        state.deal_hole('QcQd')
        state.deal_hole('JcJd')
        state.deal_hole('TcTd')
        state.deal_hole('9c9d')
        self.assertRaises(ValueError, state.complete_bet_or_raise_to)
        state.check_or_call()
        state.burn_card('??')
        state.deal_board('QhQs2c')
        state.burn_card('??')
        state.deal_board('3c')
        state.burn_card('??')
        state.deal_board('4c')
        self.assertFalse(state.status)
        self.assertEqual(state.stacks, [0, 0, 18, 0, 0, 0])

        state = NoLimitTexasHoldem.create_state(
            (
                Automation.ANTE_POSTING,
                Automation.BET_COLLECTION,
                Automation.BLIND_OR_STRADDLE_POSTING,
                Automation.HOLE_CARDS_SHOWING_OR_MUCKING,
                Automation.HAND_KILLING,
                Automation.CHIPS_PUSHING,
                Automation.CHIPS_PULLING,
            ),
            True,
            0,
            (1, 2),
            2,
            (200, 2, 200),
            3,
        )

        state.deal_hole('AcAd')
        state.deal_hole('KcKd')
        state.deal_hole('QcQd')
        state.complete_bet_or_raise_to(100)
        state.check_or_call()
        state.burn_card('??')
        state.deal_board('QhQs2c')
        state.complete_bet_or_raise_to(100)
        state.check_or_call()
        state.burn_card('??')
        state.deal_board('3c')
        state.burn_card('??')
        state.deal_board('4c')
        self.assertFalse(state.status)
        self.assertEqual(state.stacks, [0, 0, 402])

        state = NoLimitTexasHoldem.create_state(
            (
                Automation.ANTE_POSTING,
                Automation.BET_COLLECTION,
                Automation.BLIND_OR_STRADDLE_POSTING,
                Automation.HOLE_CARDS_SHOWING_OR_MUCKING,
                Automation.HAND_KILLING,
                Automation.CHIPS_PUSHING,
                Automation.CHIPS_PULLING,
            ),
            True,
            0,
            (1, 2),
            2,
            (2, 200, 200),
            3,
        )

        state.deal_hole('AcAd')
        state.deal_hole('KcKd')
        state.deal_hole('QcQd')
        state.complete_bet_or_raise_to(100)
        self.assertRaises(ValueError, state.complete_bet_or_raise_to)
        state.check_or_call()
        state.check_or_call()
        state.burn_card('??')
        state.deal_board('QhQs2c')
        state.complete_bet_or_raise_to(100)
        state.check_or_call()
        state.burn_card('??')
        state.deal_board('3c')
        state.burn_card('??')
        state.deal_board('4c')
        self.assertFalse(state.status)
        self.assertEqual(state.stacks, [0, 0, 402])

        state = NoLimitShortDeckHoldem.create_state(
            (
                Automation.ANTE_POSTING,
                Automation.BET_COLLECTION,
                Automation.BLIND_OR_STRADDLE_POSTING,
                Automation.HOLE_CARDS_SHOWING_OR_MUCKING,
                Automation.HAND_KILLING,
                Automation.CHIPS_PUSHING,
                Automation.CHIPS_PULLING,
            ),
            True,
            2,
            {-1: 2},
            2,
            (2, 2, 2, 2, 2, 6),
            6,
        )

        state.deal_hole('AcAd')
        state.deal_hole('KcKd')
        state.deal_hole('QcQd')
        state.deal_hole('JcJd')
        state.deal_hole('TcTd')
        state.deal_hole('9c9d')
        state.burn_card('??')
        state.deal_board('QhQs6c')
        state.burn_card('??')
        state.deal_board('7c')
        state.burn_card('??')
        state.deal_board('8c')
        self.assertFalse(state.status)
        self.assertEqual(state.stacks, [0, 0, 12, 0, 0, 4])

        state = NoLimitShortDeckHoldem.create_state(
            (
                Automation.ANTE_POSTING,
                Automation.BET_COLLECTION,
                Automation.BLIND_OR_STRADDLE_POSTING,
                Automation.HOLE_CARDS_SHOWING_OR_MUCKING,
                Automation.HAND_KILLING,
                Automation.CHIPS_PUSHING,
                Automation.CHIPS_PULLING,
            ),
            True,
            2,
            {-1: 2},
            2,
            (2, 2, 3, 3, 2, 6),
            6,
        )

        state.deal_hole('AcAd')
        state.deal_hole('KcKd')
        state.deal_hole('QcQd')
        state.deal_hole('JcJd')
        state.deal_hole('TcTd')
        state.deal_hole('9c9d')
        state.check_or_call()
        state.check_or_call()
        state.burn_card('??')
        state.deal_board('QhQs6c')
        state.burn_card('??')
        state.deal_board('7c')
        state.burn_card('??')
        state.deal_board('8c')
        self.assertFalse(state.status)
        self.assertEqual(state.stacks, [0, 0, 15, 0, 0, 3])

        state = NoLimitShortDeckHoldem.create_state(
            (
                Automation.ANTE_POSTING,
                Automation.BET_COLLECTION,
                Automation.BLIND_OR_STRADDLE_POSTING,
                Automation.HOLE_CARDS_SHOWING_OR_MUCKING,
                Automation.HAND_KILLING,
                Automation.CHIPS_PUSHING,
                Automation.CHIPS_PULLING,
            ),
            True,
            2,
            {-1: 2},
            2,
            (2, 2, 3, 3, 2, 6),
            6,
        )

        state.deal_hole('AcAd')
        state.deal_hole('KcKd')
        state.deal_hole('QcQd')
        state.deal_hole('JcJd')
        state.deal_hole('TcTd')
        state.deal_hole('9c9d')
        state.check_or_call()
        state.fold()
        state.burn_card('??')
        state.deal_board('QhQs6c')
        state.burn_card('??')
        state.deal_board('7c')
        state.burn_card('??')
        state.deal_board('8c')
        self.assertFalse(state.status)
        self.assertEqual(state.stacks, [0, 0, 14, 1, 0, 3])

        state = NoLimitShortDeckHoldem.create_state(
            (
                Automation.ANTE_POSTING,
                Automation.BET_COLLECTION,
                Automation.BLIND_OR_STRADDLE_POSTING,
                Automation.HOLE_CARDS_SHOWING_OR_MUCKING,
                Automation.HAND_KILLING,
                Automation.CHIPS_PUSHING,
                Automation.CHIPS_PULLING,
            ),
            True,
            2,
            {-1: 2},
            2,
            (2, 2, 5, 3, 2, 6),
            6,
        )

        state.deal_hole('AcAd')
        state.deal_hole('KcKd')
        state.deal_hole('QcQd')
        state.deal_hole('JcJd')
        state.deal_hole('TcTd')
        state.deal_hole('9c9d')
        state.check_or_call()
        state.fold()
        state.complete_bet_or_raise_to()
        state.check_or_call()
        state.burn_card('??')
        state.deal_board('QhQs6c')
        state.burn_card('??')
        state.deal_board('7c')
        state.burn_card('??')
        state.deal_board('8c')
        self.assertFalse(state.status)
        self.assertEqual(state.stacks, [0, 0, 18, 1, 0, 1])

        state = FixedLimitOmahaHoldemHighLowSplitEightOrBetter.create_state(
            (
                Automation.ANTE_POSTING,
                Automation.BET_COLLECTION,
                Automation.BLIND_OR_STRADDLE_POSTING,
                Automation.HOLE_CARDS_SHOWING_OR_MUCKING,
                Automation.HAND_KILLING,
                Automation.CHIPS_PUSHING,
                Automation.CHIPS_PULLING,
            ),
            True,
            100,
            (100, 200),
            200,
            400,
            150,
            2,
        )

        state.deal_hole('AcAdAhAs')
        state.deal_hole('KcKdKhKs')
        state.burn_card('??')
        state.deal_board('JcJdJh')
        state.burn_card('??')
        state.deal_board('Js')
        state.burn_card('??')
        state.deal_board('Tc')
        self.assertFalse(state.status)
        self.assertEqual(state.stacks, [300, 0])

        state = FixedLimitSevenCardStud.create_state(
            (
                Automation.ANTE_POSTING,
                Automation.BET_COLLECTION,
                Automation.BLIND_OR_STRADDLE_POSTING,
                Automation.HOLE_CARDS_SHOWING_OR_MUCKING,
                Automation.HAND_KILLING,
                Automation.CHIPS_PUSHING,
                Automation.CHIPS_PULLING,
            ),
            True,
            0,
            50,
            200,
            400,
            (200, 150, 175),
            3,
        )

        state.deal_hole('AcAdAh')
        state.deal_hole('KcKdKh')
        state.deal_hole('QcQdQh')
        state.complete_bet_or_raise_to()
        state.check_or_call()
        state.check_or_call()
        state.burn_card('??')
        state.deal_hole('As')
        state.deal_hole('Ks')
        state.deal_hole('Qs')
        state.burn_card('??')
        state.deal_hole('2c')
        state.deal_hole('2d')
        state.deal_hole('2h')
        state.burn_card('??')
        state.deal_hole('3c')
        state.deal_hole('3d')
        state.deal_hole('3h')
        state.burn_card('??')
        state.deal_hole('4c')
        state.deal_hole('4d')
        state.deal_hole('4h')
        self.assertFalse(state.status)
        self.assertEqual(state.stacks, [525, 0, 0])

        state = FixedLimitSevenCardStud.create_state(
            (
                Automation.ANTE_POSTING,
                Automation.BET_COLLECTION,
                Automation.BLIND_OR_STRADDLE_POSTING,
                Automation.HOLE_CARDS_SHOWING_OR_MUCKING,
                Automation.HAND_KILLING,
                Automation.CHIPS_PUSHING,
                Automation.CHIPS_PULLING,
            ),
            True,
            0,
            50,
            200,
            400,
            (200, 150, 175),
            3,
        )

        state.deal_hole('AcAdAh')
        state.deal_hole('KcKdKh')
        state.deal_hole('QcQdQh')
        state.post_bring_in()
        state.check_or_call()
        state.complete_bet_or_raise_to()
        state.check_or_call()
        state.complete_bet_or_raise_to()
        state.check_or_call()
        state.burn_card('??')
        state.deal_hole('As')
        state.deal_hole('Ks')
        state.deal_hole('Qs')
        state.burn_card('??')
        state.deal_hole('2c')
        state.deal_hole('2d')
        state.deal_hole('2h')
        state.burn_card('??')
        state.deal_hole('3c')
        state.deal_hole('3d')
        state.deal_hole('3h')
        state.burn_card('??')
        state.deal_hole('4c')
        state.deal_hole('4d')
        state.deal_hole('4h')
        self.assertFalse(state.status)
        self.assertEqual(state.stacks, [525, 0, 0])

    def test_automated_dealing(self) -> None:
        state = FixedLimitDeuceToSevenLowballTripleDraw.create_state(
            (
                Automation.ANTE_POSTING,
                Automation.BET_COLLECTION,
                Automation.BLIND_OR_STRADDLE_POSTING,
                Automation.HOLE_DEALING,
                Automation.BOARD_DEALING,
                Automation.HOLE_CARDS_SHOWING_OR_MUCKING,
                Automation.HAND_KILLING,
                Automation.CHIPS_PUSHING,
                Automation.CHIPS_PULLING,
            ),
            True,
            0,
            (1, 2),
            2,
            4,
            200,
            6,
        )

        state.check_or_call()
        state.check_or_call()
        state.check_or_call()
        state.check_or_call()
        state.check_or_call()
        state.check_or_call()

        state.stand_pat_or_discard(state.hole_cards[0])
        state.stand_pat_or_discard(state.hole_cards[1])
        state.stand_pat_or_discard(state.hole_cards[2])
        state.stand_pat_or_discard(state.hole_cards[3])
        state.stand_pat_or_discard(state.hole_cards[4])
        state.stand_pat_or_discard(state.hole_cards[5])
        state.burn_card('??')
        state.check_or_call()

    def test_reshuffling(self) -> None:
        state = FixedLimitDeuceToSevenLowballTripleDraw.create_state(
            (
                Automation.ANTE_POSTING,
                Automation.BET_COLLECTION,
                Automation.BLIND_OR_STRADDLE_POSTING,
                Automation.CARD_BURNING,
                Automation.HOLE_DEALING,
                Automation.BOARD_DEALING,
                Automation.HOLE_CARDS_SHOWING_OR_MUCKING,
                Automation.HAND_KILLING,
                Automation.CHIPS_PUSHING,
                Automation.CHIPS_PULLING,
            ),
            True,
            0,
            (1, 2),
            2,
            4,
            200,
            6,
        )

        state.check_or_call()
        state.check_or_call()
        state.check_or_call()
        state.check_or_call()
        state.check_or_call()
        state.check_or_call()

        state.stand_pat_or_discard(state.hole_cards[0])
        state.stand_pat_or_discard(state.hole_cards[1])
        state.stand_pat_or_discard(state.hole_cards[2])
        state.stand_pat_or_discard(state.hole_cards[3])
        state.stand_pat_or_discard(state.hole_cards[4])
        state.stand_pat_or_discard(state.hole_cards[5])
        state.check_or_call()
        state.check_or_call()
        state.check_or_call()
        state.check_or_call()
        state.check_or_call()
        state.check_or_call()

        state.stand_pat_or_discard(state.hole_cards[0])
        state.stand_pat_or_discard(state.hole_cards[1])
        state.stand_pat_or_discard(state.hole_cards[2])
        state.stand_pat_or_discard(state.hole_cards[3])
        state.stand_pat_or_discard(state.hole_cards[4])
        state.stand_pat_or_discard(state.hole_cards[5])
        state.check_or_call()
        state.check_or_call()
        state.check_or_call()
        state.check_or_call()
        state.check_or_call()
        state.check_or_call()

        state.stand_pat_or_discard(state.hole_cards[0])
        state.stand_pat_or_discard(state.hole_cards[1])
        state.stand_pat_or_discard(state.hole_cards[2])
        state.stand_pat_or_discard(state.hole_cards[3])
        state.stand_pat_or_discard(state.hole_cards[4])
        state.stand_pat_or_discard(state.hole_cards[5])
        state.check_or_call()
        state.check_or_call()
        state.check_or_call()
        state.check_or_call()
        state.check_or_call()
        state.check_or_call()

    def test_hole_to_board_dealing(self) -> None:
        state = FixedLimitRazz.create_state(
            (
                Automation.ANTE_POSTING,
                Automation.BET_COLLECTION,
                Automation.BLIND_OR_STRADDLE_POSTING,
                Automation.CARD_BURNING,
                Automation.HOLE_CARDS_SHOWING_OR_MUCKING,
                Automation.HAND_KILLING,
                Automation.CHIPS_PUSHING,
                Automation.CHIPS_PULLING,
            ),
            True,
            0,
            1,
            2,
            4,
            200,
            9,
        )

        for _ in range(3 * state.player_count):
            state.deal_hole()

        state.complete_bet_or_raise_to()

        for _ in range(state.player_count - 1):
            state.check_or_call()

        for _ in range(state.player_count):
            state.deal_hole()

        for _ in range(state.player_count):
            state.check_or_call()

        for _ in range(state.player_count):
            state.deal_hole()

        for _ in range(state.player_count):
            state.check_or_call()

        self.assertFalse(state.can_deal_hole())
        self.assertRaises(ValueError, state.deal_hole)
        self.assertTrue(state.can_deal_board())
        state.deal_board()

        for _ in range(state.player_count):
            state.check_or_call()

        self.assertFalse(state.can_deal_hole())
        self.assertRaises(ValueError, state.deal_hole)
        self.assertTrue(state.can_deal_board())
        state.deal_board()

        for _ in range(state.player_count):
            state.check_or_call()

    def test_preflop_opener(self) -> None:

        def create_state(
                forced_bets: ValuesLike,
                player_count: int,
        ) -> State:
            return NoLimitTexasHoldem.create_state(
                tuple(Automation),
                False,
                0,
                forced_bets,
                2,
                200,
                player_count,
            )

        self.assertEqual(create_state((1, 2), 2).actor_index, 1)
        self.assertEqual(create_state((1, 2), 3).actor_index, 2)
        self.assertEqual(create_state((1, 2), 6).actor_index, 2)
        self.assertEqual(create_state((1, 2, 4), 6).actor_index, 3)
        self.assertEqual(create_state((1, 2, 4, 8), 6).actor_index, 4)
        self.assertEqual(create_state((1, 2, 0, 0, 0, 4), 6).actor_index, 0)
        self.assertEqual(create_state((1, 2, 4, 0, 0, 8), 6).actor_index, 0)

    def test_chips_pushing(self) -> None:
        state = NoLimitTexasHoldem.create_state(
            (
                Automation.ANTE_POSTING,
                Automation.BET_COLLECTION,
                Automation.BLIND_OR_STRADDLE_POSTING,
                Automation.HOLE_CARDS_SHOWING_OR_MUCKING,
                Automation.HAND_KILLING,
                Automation.CHIPS_PULLING,
            ),
            True,
            0,
            (1, 2),
            2,
            (100, 10),
            2,
            rake=partial(rake, rake=0.1),
        )

        state.deal_hole('2c3c')
        state.deal_hole('2d3d')
        state.complete_bet_or_raise_to(10)
        state.check_or_call()
        state.burn_card('??')
        state.deal_board('TsJsQs')
        state.burn_card('??')
        state.deal_board('Ks')
        state.burn_card('??')
        state.deal_board('As')
        state.push_chips()
        self.assertFalse(state.status)
        self.assertEqual(state.stacks, [99, 9])

        state = NoLimitTexasHoldem.create_state(
            (
                Automation.ANTE_POSTING,
                Automation.BET_COLLECTION,
                Automation.BLIND_OR_STRADDLE_POSTING,
                Automation.HOLE_CARDS_SHOWING_OR_MUCKING,
                Automation.HAND_KILLING,
                Automation.CHIPS_PULLING,
            ),
            True,
            0,
            (1, 2),
            2,
            (100, 10, 1000),
            3,
            rake=partial(rake, rake=0.1),
        )

        state.deal_hole('2c3c')
        state.deal_hole('2d3d')
        state.deal_hole('2h3h')
        state.complete_bet_or_raise_to(1000)
        state.check_or_call()
        state.check_or_call()
        state.burn_card('??')
        state.deal_board('TsJsQs')
        state.burn_card('??')
        state.deal_board('Ks')
        state.burn_card('??')
        state.deal_board('As')
        state.push_chips()
        state.push_chips()
        self.assertFalse(state.status)
        self.assertEqual(state.stacks, [90, 9, 990])

        state = NoLimitTexasHoldem.create_state(
            (
                Automation.ANTE_POSTING,
                Automation.BET_COLLECTION,
                Automation.BLIND_OR_STRADDLE_POSTING,
                Automation.HOLE_CARDS_SHOWING_OR_MUCKING,
                Automation.HAND_KILLING,
                Automation.CHIPS_PULLING,
            ),
            True,
            0,
            (1, 2),
            2,
            (100, 10, 10000, 1000),
            4,
            rake=partial(rake, rake=0.1),
        )

        state.deal_hole('2c3c')
        state.deal_hole('2d3d')
        state.deal_hole('2h3h')
        state.deal_hole('2s3s')
        state.complete_bet_or_raise_to(10000)
        state.check_or_call()
        state.check_or_call()
        state.check_or_call()
        state.burn_card('??')
        state.deal_board('TsJsQs')
        state.burn_card('??')
        state.deal_board('Ks')
        state.burn_card('??')
        state.deal_board('As')
        state.push_chips()
        state.push_chips()
        state.push_chips()
        self.assertFalse(state.status)
        self.assertEqual(state.stacks, [90, 9, 9900, 900])

        state = NoLimitTexasHoldem.create_state(
            (
                Automation.ANTE_POSTING,
                Automation.BET_COLLECTION,
                Automation.BLIND_OR_STRADDLE_POSTING,
                Automation.HOLE_CARDS_SHOWING_OR_MUCKING,
                Automation.HAND_KILLING,
                Automation.CHIPS_PULLING,
            ),
            True,
            0,
            (1, 2),
            2,
            (100, 10, 100000, 10000, 1000),
            5,
            rake=partial(rake, rake=0.1),
        )

        state.deal_hole('2c3c')
        state.deal_hole('2d3d')
        state.deal_hole('2h3h')
        state.deal_hole('2s3s')
        state.deal_hole('4c5c')
        state.complete_bet_or_raise_to(100000)
        state.check_or_call()
        state.check_or_call()
        state.check_or_call()
        state.check_or_call()
        state.burn_card('??')
        state.deal_board('TsJsQs')
        state.burn_card('??')
        state.deal_board('Ks')
        state.burn_card('??')
        state.deal_board('As')
        state.push_chips()
        state.push_chips()
        state.push_chips()
        state.push_chips()
        self.assertFalse(state.status)
        self.assertEqual(state.stacks, [90, 9, 99000, 9000, 900])


if __name__ == '__main__':
    main()  # pragma: no cover



""":mod:`pokerkit` is the top-level package for the PokerKit library.

All poker tools are imported here.
"""

__all__ = (
    'AntePosting',
    'Automation',
    'BadugiHand',
    'BadugiLookup',
    'BetCollection',
    'BettingStructure',
    'BlindOrStraddlePosting',
    'BoardCombinationHand',
    'BoardDealing',
    'BringInPosting',
    'Card',
    'CardBurning',
    'CardsLike',
    'CheckingOrCalling',
    'ChipsPulling',
    'ChipsPushing',
    'clean_values',
    'CombinationHand',
    'CompletionBettingOrRaisingTo',
    'Deck',
    'DeuceToSevenLowballMixin',
    'divmod',
    'Draw',
    'EightOrBetterLookup',
    'EightOrBetterLowHand',
    'Entry',
    'filter_none',
    'FixedLimitBadugi',
    'FixedLimitDeuceToSevenLowballTripleDraw',
    'FixedLimitOmahaHoldemHighLowSplitEightOrBetter',
    'FixedLimitPokerMixin',
    'FixedLimitRazz',
    'FixedLimitSevenCardStud',
    'FixedLimitSevenCardStudHighLowSplitEightOrBetter',
    'FixedLimitTexasHoldem',
    'Folding',
    'GreekHoldemHand',
    'Hand',
    'HandHistory',
    'HandKilling',
    'Holdem',
    'HoleBoardCombinationHand',
    'HoleCardsShowingOrMucking',
    'HoleDealing',
    'KuhnPokerHand',
    'KuhnPokerLookup',
    'Label',
    'Lookup',
    'max_or_none',
    'min_or_none',
    'NoLimitDeuceToSevenLowballSingleDraw',
    'NoLimitPokerMixin',
    'NoLimitShortDeckHoldem',
    'NoLimitTexasHoldem',
    'OmahaEightOrBetterLowHand',
    'OmahaHoldemHand',
    'OmahaHoldemMixin',
    'Opening',
    'Operation',
    'parse_action',
    'parse_value',
    'Poker',
    'Pot',
    'PotLimitOmahaHoldem',
    'PotLimitPokerMixin',
    'rake',
    'Rank',
    'RankOrder',
    'RegularLookup',
    'RegularLowHand',
    'SevenCardStud',
    'ShortDeckHoldemHand',
    'ShortDeckHoldemLookup',
    'shuffled',
    'SingleDraw',
    'StandardHand',
    'StandardHighHand',
    'StandardLookup',
    'StandardLowHand',
    'StandingPatOrDiscarding',
    'State',
    'Street',
    'Suit',
    'TexasHoldemMixin',
    'TripleDraw',
    'UnfixedLimitHoldem',
    'ValuesLike',
)

from pokerkit.games import (
    DeuceToSevenLowballMixin,
    Draw,
    FixedLimitBadugi,
    FixedLimitDeuceToSevenLowballTripleDraw,
    FixedLimitOmahaHoldemHighLowSplitEightOrBetter,
    FixedLimitPokerMixin,
    FixedLimitRazz,
    FixedLimitSevenCardStud,
    FixedLimitSevenCardStudHighLowSplitEightOrBetter,
    FixedLimitTexasHoldem,
    Holdem,
    NoLimitDeuceToSevenLowballSingleDraw,
    NoLimitPokerMixin,
    NoLimitShortDeckHoldem,
    NoLimitTexasHoldem,
    OmahaHoldemMixin,
    Poker,
    PotLimitOmahaHoldem,
    PotLimitPokerMixin,
    SevenCardStud,
    SingleDraw,
    TexasHoldemMixin,
    TripleDraw,
    UnfixedLimitHoldem,
)
from pokerkit.hands import (
    BadugiHand,
    BoardCombinationHand,
    CombinationHand,
    EightOrBetterLowHand,
    GreekHoldemHand,
    Hand,
    HoleBoardCombinationHand,
    KuhnPokerHand,
    OmahaEightOrBetterLowHand,
    OmahaHoldemHand,
    RegularLowHand,
    ShortDeckHoldemHand,
    StandardHand,
    StandardHighHand,
    StandardLowHand,
)
from pokerkit.lookups import (
    BadugiLookup,
    EightOrBetterLookup,
    Entry,
    KuhnPokerLookup,
    Label,
    Lookup,
    RegularLookup,
    ShortDeckHoldemLookup,
    StandardLookup,
)
from pokerkit.notation import HandHistory, parse_action
from pokerkit.state import (
    AntePosting,
    Automation,
    BetCollection,
    BettingStructure,
    BlindOrStraddlePosting,
    BoardDealing,
    BringInPosting,
    CardBurning,
    CheckingOrCalling,
    ChipsPulling,
    ChipsPushing,
    CompletionBettingOrRaisingTo,
    Folding,
    HandKilling,
    HoleCardsShowingOrMucking,
    HoleDealing,
    Opening,
    Operation,
    Pot,
    StandingPatOrDiscarding,
    State,
    Street,
)
from pokerkit.utilities import (
    Card,
    CardsLike,
    clean_values,
    Deck,
    divmod,
    filter_none,
    max_or_none,
    min_or_none,
    parse_value,
    rake,
    Rank,
    RankOrder,
    shuffled,
    Suit,
    ValuesLike,
)



""":mod:`pokerkit.notation` implements classes related to poker
notations.
"""

from __future__ import annotations

from collections.abc import Callable, Iterable, Iterator
from collections import defaultdict, deque
from dataclasses import asdict, dataclass, fields, KW_ONLY
from operator import itemgetter
from tomllib import loads as loads_toml
from typing import Any, ClassVar, BinaryIO
from warnings import warn
import datetime

from pokerkit.state import (
    Automation,
    BoardDealing,
    BringInPosting,
    CheckingOrCalling,
    CompletionBettingOrRaisingTo,
    Folding,
    HoleCardsShowingOrMucking,
    HoleDealing,
    StandingPatOrDiscarding,
    State,
)
from pokerkit.games import (
    FixedLimitBadugi,
    FixedLimitDeuceToSevenLowballTripleDraw,
    FixedLimitOmahaHoldemHighLowSplitEightOrBetter,
    FixedLimitRazz,
    FixedLimitSevenCardStud,
    FixedLimitSevenCardStudHighLowSplitEightOrBetter,
    FixedLimitTexasHoldem,
    NoLimitDeuceToSevenLowballSingleDraw,
    NoLimitShortDeckHoldem,
    NoLimitTexasHoldem,
    Poker,
    PotLimitOmahaHoldem,
)
from pokerkit.utilities import Card, divmod, parse_value


@dataclass
class HandHistory(Iterable[State]):
    game_types: ClassVar[dict[str, type[Poker]]] = {
        'FT': FixedLimitTexasHoldem,
        'NT': NoLimitTexasHoldem,
        'NS': NoLimitShortDeckHoldem,
        'PO': PotLimitOmahaHoldem,
        'FO/8': FixedLimitOmahaHoldemHighLowSplitEightOrBetter,
        'F7S': FixedLimitSevenCardStud,
        'F7S/8': FixedLimitSevenCardStudHighLowSplitEightOrBetter,
        'FR': FixedLimitRazz,
        'N2L1D': NoLimitDeuceToSevenLowballSingleDraw,
        'F2L3D': FixedLimitDeuceToSevenLowballTripleDraw,
        'FB': FixedLimitBadugi,
    }
    """The game codes and the corresponding game types."""
    variants: ClassVar[dict[type[Poker], str]] = dict(
        zip(game_types.values(), game_types.keys()),
    )
    """The game types and the corresponding game codes."""
    required_field_names: ClassVar[dict[str, tuple[str, ...]]] = {
        'FT': (
            'variant',
            'antes',
            'blinds_or_straddles',
            'small_bet',
            'big_bet',
            'starting_stacks',
            'actions',
        ),
        'NT': (
            'variant',
            'antes',
            'blinds_or_straddles',
            'min_bet',
            'starting_stacks',
            'actions',
        ),
        'NS': (
            'variant',
            'antes',
            'blinds_or_straddles',
            'min_bet',
            'starting_stacks',
            'actions',
        ),
        'PO': (
            'variant',
            'antes',
            'blinds_or_straddles',
            'min_bet',
            'starting_stacks',
            'actions',
        ),
        'FO/8': (
            'variant',
            'antes',
            'blinds_or_straddles',
            'small_bet',
            'big_bet',
            'starting_stacks',
            'actions',
        ),
        'F7S': (
            'variant',
            'antes',
            'bring_in',
            'small_bet',
            'big_bet',
            'starting_stacks',
            'actions',
        ),
        'F7S/8': (
            'variant',
            'antes',
            'bring_in',
            'small_bet',
            'big_bet',
            'starting_stacks',
            'actions',
        ),
        'FR': (
            'variant',
            'antes',
            'bring_in',
            'small_bet',
            'big_bet',
            'starting_stacks',
            'actions',
        ),
        'N2L1D': (
            'variant',
            'antes',
            'blinds_or_straddles',
            'min_bet',
            'starting_stacks',
            'actions',
        ),
        'F2L3D': (
            'variant',
            'antes',
            'blinds_or_straddles',
            'small_bet',
            'big_bet',
            'starting_stacks',
            'actions',
        ),
        'FB': (
            'variant',
            'antes',
            'blinds_or_straddles',
            'small_bet',
            'big_bet',
            'starting_stacks',
            'actions',
        ),
    }
    """The required field names."""
    optional_field_names: ClassVar[tuple[str, ...]] = (
        'author',
        'event',
        'url',
        'address',
        'city',
        'region',
        'postal_code',
        'country',
        'time',
        'time_zone',
        'day',
        'month',
        'year',
        'hand',
        'level',
        'seats',
        'seat_count',
        'table',
        'players',
        'finishing_stacks',
        'currency',
        'ante_trimming_status',
        'time_limit',
        'time_banks',
    )
    """The optional field names."""
    _: KW_ONLY
    variant: str
    """The variant name."""
    ante_trimming_status: bool = False
    """The ante trimming status."""
    antes: list[int]
    """The antes."""
    blinds_or_straddles: list[int] | None = None
    """The blinds or straddles."""
    bring_in: int | None = None
    """The bring-in."""
    small_bet: int | None = None
    """The small bet."""
    big_bet: int | None = None
    """The big bet."""
    min_bet: int | None = None
    """The minimum bet."""
    starting_stacks: list[int]
    """The starting stacks."""
    actions: list[str]
    """The actions."""
    author: str | None = None
    """The author."""
    event: str | None = None
    """The event."""
    url: str | None = None
    """The url."""
    address: str | None = None
    """The address."""
    city: str | None = None
    """The city."""
    region: str | None = None
    """The region."""
    postal_code: str | None = None
    """The postal code."""
    country: str | None = None
    """The country."""
    time: datetime.time | None = None
    """The time."""
    time_zone: str | None = None
    """The time zone."""
    day: int | None = None
    """The day."""
    month: int | None = None
    """The month."""
    year: int | None = None
    """The year."""
    hand: int | None = None
    """The hand number."""
    level: int | None = None
    """The level."""
    seats: list[int] | None = None
    """The seat numbers."""
    seat_count: int | None = None
    """The number of seats."""
    table: int | None = None
    """The table number."""
    players: list[str] | None = None
    """The player names."""
    finishing_stacks: list[int] | None = None
    """The finishing stacks."""
    currency: str | None = None
    """The currency."""
    time_limit: int | None = None
    """The time limit."""
    time_banks: list[int] | None = None
    """The time banks."""
    automations: tuple[Automation, ...] = (
        Automation.ANTE_POSTING,
        Automation.BET_COLLECTION,
        Automation.BLIND_OR_STRADDLE_POSTING,
        Automation.CARD_BURNING,
        Automation.HAND_KILLING,
        Automation.CHIPS_PUSHING,
        Automation.CHIPS_PULLING,
    )
    """The automations."""
    divmod: Callable[[int, int], tuple[int, int]] = divmod
    """The divmod function."""
    parse_value: Callable[[str], int] = parse_value
    """The value parsing function."""

    @classmethod
    def _filter_non_fields(cls, **kwargs: Any) -> dict[str, Any]:
        field_names = {field.name for field in fields(cls)}

        for key in tuple(kwargs.keys()):
            if key not in field_names and not key.startswith('_'):
                warn(f'unexpected field \'{key}\'')
                kwargs.pop(key)

        return kwargs

    @classmethod
    def loads(cls, s: str, **kwargs: Any) -> HandHistory:
        """Load PHH from ``str`` object.

        :param s: The ``str`` object.
        :param kwargs: The metadata.
        :return: The hand history object.
        """
        return cls(**cls._filter_non_fields(**kwargs | loads_toml(s)))

    @classmethod
    def load(cls, fp: BinaryIO, **kwargs: Any) -> HandHistory:
        """Load PHH from a file pointer.

        :param fp: The file pointer.
        :param kwargs: The metadata.
        :return: The hand history object.
        """
        return cls.loads(fp.read().decode(), **kwargs)

    @classmethod
    def from_game_state(
            cls,
            game: Poker,
            state: State,
            compression_status: bool = True,
            **kwargs: Any,
    ) -> HandHistory:
        """Create a hand history from game state.

        :param game: The game.
        :param state: The state.
        :param compression_status: The compression status.
        :param kwargs: The metadata.
        :return: The hand history.
        """
        variant = cls.variants[type(game)]
        actions = []
        cards = defaultdict[int, list[Card]](list)

        for operation in state.operations:
            action: str | None

            if (
                    not compression_status
                    or not isinstance(operation, HoleDealing)
            ):
                for player_index in cards:
                    if cards[player_index]:
                        action = (
                            f'd dh p{player_index + 1} '
                            + ''.join(map(repr, cards[player_index]))
                        )

                        actions.append(action.strip())
                        cards[player_index].clear()

            if isinstance(operation, BoardDealing):
                action = 'd db ' + ''.join(map(repr, operation.cards))
            elif isinstance(operation, HoleDealing):
                cards[operation.player_index].extend(operation.cards)

                action = None
            elif isinstance(operation, StandingPatOrDiscarding):
                action = (
                    f'p{operation.player_index + 1} sd '
                    + ''.join(map(repr, operation.cards))
                )
            elif isinstance(operation, BringInPosting):
                action = f'p{operation.player_index + 1} pb'
            elif isinstance(operation, Folding):
                action = f'p{operation.player_index + 1} f'
            elif isinstance(operation, CheckingOrCalling):
                action = f'p{operation.player_index + 1} cc'
            elif isinstance(operation, CompletionBettingOrRaisingTo):
                action = (
                    f'p{operation.player_index + 1} cbr {operation.amount}'
                )
            elif isinstance(operation, HoleCardsShowingOrMucking):
                action = (
                    f'p{operation.player_index + 1} sm '
                    + ''.join(map(repr, operation.hole_cards))
                )
            else:
                action = None

            if action is not None:
                actions.append(action.strip())

        assert not any(cards.values())

        kwargs.setdefault('variant', variant)
        kwargs.setdefault('actions', actions)
        kwargs.setdefault('starting_stacks', list(state.starting_stacks))

        field_names = {field.name for field in fields(cls)}

        for name in cls.required_field_names[variant]:
            if name not in field_names:
                continue

            try:
                attribute = getattr(game, name)
            except AttributeError:
                try:
                    attribute = getattr(state, name)
                except AttributeError:
                    continue

            if isinstance(attribute, Iterable):
                attribute = list(attribute)

            kwargs.setdefault(name, attribute)

        return HandHistory(**cls._filter_non_fields(**kwargs))

    def __iter__(self) -> Iterator[State]:
        yield from map(itemgetter(0), self.iter_state_actions())

    def iter_state_actions(self) -> Iterator[tuple[State, str | None]]:
        """Iterate through state-actions.

        :return: The state actions.
        """
        state = self.create_state()
        actions = deque(self.actions)

        yield state, None

        while state.status:
            action = None

            if state.can_post_ante():
                state.post_ante()
            elif state.can_collect_bets():
                state.collect_bets()
            elif state.can_post_blind_or_straddle():
                state.post_blind_or_straddle()
            elif state.can_burn_card():
                state.burn_card('??')

                if Automation.CARD_BURNING in self.automations:
                    continue
            elif state.can_kill_hand():
                state.kill_hand()
            elif state.can_push_chips():
                state.push_chips()
            elif state.can_pull_chips():
                state.pull_chips()
            else:
                action = actions.popleft()

                parse_action(state, action, self.parse_value)

            yield state, action

    @property
    def game_type(self) -> type[Poker]:
        """Return the game type.

        :return: The game type.
        """
        return self.game_types[self.variant]

    def create_game(self) -> Poker:
        """Create the game.

        :return: The game.
        """
        automations = list(self.automations)

        if Automation.CARD_BURNING in self.automations:
            automations.remove(Automation.CARD_BURNING)

        kwargs: dict[str, Any] = {
            'automations': tuple(automations),
            'divmod': self.divmod,
            'ante_trimming_status': self.ante_trimming_status,
        }

        for name in self.required_field_names[self.variant]:
            if name == 'antes' or name == 'blinds_or_straddles':
                key = f'raw_{name}'
            else:
                key = name

            kwargs[key] = getattr(self, name)

        kwargs.pop('variant')
        kwargs.pop('starting_stacks')
        kwargs.pop('actions')

        return self.game_type(**kwargs)

    def create_state(self) -> State:
        """Create the initial state.

        :return: The initial state.
        """
        return self.create_game()(
            self.starting_stacks,
            len(self.starting_stacks),
        )

    def dumps(self) -> str:
        """Dump PHH as a ``str`` object.

        :return: a ``str`` object.
        """

        def clean(value: Any) -> str:
            cleaned_value: str

            if isinstance(value, bool):
                cleaned_value = repr(value).lower()
            elif isinstance(value, datetime.time):
                cleaned_value = str(value)
            else:
                cleaned_value = repr(value)

            return cleaned_value

        lines = []

        for key, value in asdict(self).items():
            if (
                    (
                        key in self.required_field_names[self.variant]
                        or key in self.optional_field_names
                    )
                    and value is not None
            ):
                lines.append(f'{key} = {clean(value)}')

        return '\n'.join(lines)

    def dump(self, fp: BinaryIO) -> None:
        """Dump PHH to a file pointer.

        :param fp: The file pointer.
        :return: ``None``.
        """
        fp.write(self.dumps().encode())


def parse_action(
        state: State,
        action: str,
        parse_value: Callable[[str], int] = parse_value,
) -> None:
    """Parse the action.

    :param state: The state.
    :param action: The string action.
    :param parse_value: The value parsing function.
    :return: ``None``.
    """
    def verify_player(index: int | None) -> None:
        label, parsed_index = player[:1], int(player[1:]) - 1

        if label != 'p' or parsed_index != index:
            raise ValueError(f'invalid Player \'{player}\'')

    words = action.split()

    if '#' in words:
        words = words[:words.index('#')]

    match words:
        case 'd', 'db', cards:
            state.deal_board(cards)
        case 'd', 'dh', player, cards:
            verify_player(state.hole_dealee_index)
            state.deal_hole(cards)
        case player, 'sd':
            verify_player(state.stander_pat_or_discarder_index)
            state.stand_pat_or_discard()
        case player, 'sd', cards:
            verify_player(state.stander_pat_or_discarder_index)
            state.stand_pat_or_discard(cards)
        case player, 'pb':
            verify_player(state.actor_index)
            state.post_bring_in()
        case player, 'f':
            verify_player(state.actor_index)
            state.fold()
        case player, 'cc':
            verify_player(state.actor_index)
            state.check_or_call()
        case player, 'cbr', amount:
            verify_player(state.actor_index)
            state.complete_bet_or_raise_to(parse_value(amount))
        case player, 'sm':
            verify_player(state.showdown_index)
            state.show_or_muck_hole_cards(False)
        case player, 'sm', '-':
            verify_player(state.showdown_index)
            state.show_or_muck_hole_cards(True)
        case player, 'sm', cards:
            verify_player(state.showdown_index)
            state.show_or_muck_hole_cards(cards)
        case ():
            pass
        case _:
            raise ValueError(f'invalid action \'{action}\'')



""":mod:`pokerkit.hands` implements classes related to poker hands."""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Hashable
from functools import total_ordering
from itertools import chain, combinations
from typing import Any, ClassVar

from pokerkit.lookups import (
    BadugiLookup,
    EightOrBetterLookup,
    Entry,
    KuhnPokerLookup,
    Lookup,
    RegularLookup,
    ShortDeckHoldemLookup,
    StandardLookup,
)
from pokerkit.utilities import Card, CardsLike, RankOrder


@total_ordering
class Hand(Hashable, ABC):
    """The abstract base class for poker hands.

    Stronger hands are considered greater than weaker hands.

    >>> h0 = ShortDeckHoldemHand('6s7s8s9sTs')
    >>> h1 = ShortDeckHoldemHand('7c8c9cTcJc')
    >>> h2 = ShortDeckHoldemHand('2c2d2h2s3h')
    Traceback (most recent call last):
        ...
    ValueError: invalid hand '2c2d2h2s3h'
    >>> h0
    6s7s8s9sTs
    >>> h1
    7c8c9cTcJc
    >>> print(h0)
    Straight flush (6s7s8s9sTs)
    >>> h0 < h1
    True

    It does not make sense to compare hands of different types.

    >>> h = BadugiHand('6d7s8h9c')
    >>> h < 500
    Traceback (most recent call last):
        ...
    TypeError: '<' not supported between instances of 'BadugiHand' and 'int'

    The hands are hashable.

    >>> h0 = ShortDeckHoldemHand('6s7s8s9sTs')
    >>> h1 = ShortDeckHoldemHand('7c8c9cTcJc')
    >>> hands = {h0, h1}
    """

    lookup: ClassVar[Lookup]
    """The hand lookup."""
    low: ClassVar[bool]
    """The low status."""

    @classmethod
    @abstractmethod
    def from_game(
            cls,
            hole_cards: CardsLike,
            board_cards: CardsLike = (),
    ) -> Hand:
        """Create a poker hand from a game setting.

        In a game setting, a player uses private cards from their hole
        and the public cards from the board to make their hand.

        :param hole_cards: The hole cards.
        :param board_cards: The optional board cards.
        :return: The strongest hand from possible card combinations.
        """
        pass  # pragma: no cover

    def __init__(self, cards: CardsLike) -> None:
        self.__cards = Card.clean(cards)

        if not self.lookup.has_entry(self.cards):
            raise ValueError(f'invalid hand \'{repr(self)}\'')

    def __eq__(self, other: Any) -> bool:
        if type(self) != type(other):  # noqa: E721
            return NotImplemented

        assert isinstance(other, Hand)

        return self.entry == other.entry

    def __hash__(self) -> int:
        return hash(self.entry)

    def __lt__(self, other: Hand) -> bool:
        if type(self) != type(other):  # noqa: E721
            return NotImplemented

        assert isinstance(other, Hand)

        if self.low:
            ordering = self.entry > other.entry
        else:
            ordering = self.entry < other.entry

        return ordering

    def __repr__(self) -> str:
        return ''.join(map(repr, self.cards))

    def __str__(self) -> str:
        return f'{self.entry.label.value} ({repr(self)})'

    @property
    def cards(self) -> tuple[Card, ...]:
        """Return the cards that form this hand.

        >>> hole = 'AsAc'
        >>> board = 'Kh3sAdAh'
        >>> hand = StandardHighHand.from_game(hole, board)
        >>> hand.cards
        (As, Ac, Kh, Ad, Ah)

        :return: The cards that form this hand.
        """
        return self.__cards

    @property
    def entry(self) -> Entry:
        """Return the hand entry.

        >>> hole = 'AsAc'
        >>> board = 'Kh3sAdAh'
        >>> hand = StandardHighHand.from_game(hole, board)
        >>> hand.entry.label
        <Label.FOUR_OF_A_KIND: 'Four of a kind'>

        :return: The hand entry.
        """
        return self.lookup.get_entry(self.cards)


class CombinationHand(Hand, ABC):
    """The abstract base class for combination hands."""

    card_count: ClassVar[int]
    """The number of cards."""

    @classmethod
    def from_game(
            cls,
            hole_cards: CardsLike,
            board_cards: CardsLike = (),
    ) -> Hand:
        """Create a poker hand from a game setting.

        In a game setting, a player uses private cards from their hole
        and the public cards from the board to make their hand.

        >>> h0 = StandardHighHand.from_game('AcAdAhAsKc')
        >>> h1 = StandardHighHand('AcAdAhAsKc')
        >>> h0 == h1
        True
        >>> h0 = StandardHighHand.from_game('Ac9c', 'AhKhQhJhTh')
        >>> h1 = StandardHighHand('AhKhQhJhTh')
        >>> h0 == h1
        True

        >>> h0 = StandardLowHand.from_game('AcAdAhAsKc', '')
        >>> h1 = StandardLowHand('AcAdAhAsKc')
        >>> h0 == h1
        True
        >>> h0 = StandardLowHand.from_game('Ac9c', 'AhKhQhJhTh')
        >>> h1 = StandardLowHand('AcQhJhTh9c')
        >>> h0 == h1
        True

        >>> h0 = ShortDeckHoldemHand.from_game('AcKs', 'AhAsKcJsTs')
        >>> h1 = ShortDeckHoldemHand('AcAhAsKcKs')
        >>> h0 == h1
        True
        >>> h0 = ShortDeckHoldemHand.from_game('AcAd', '6s7cKcKd')
        >>> h1 = ShortDeckHoldemHand('AcAdKcKd7c')
        >>> h0 == h1
        True

        >>> h0 = EightOrBetterLowHand.from_game('As2s', '2c3c4c5c6c')
        >>> h1 = EightOrBetterLowHand('Ad2d3d4d5d')
        >>> h0 == h1
        True

        >>> h0 = RegularLowHand.from_game('AcAd', 'AhAsKcQdQh')
        >>> h1 = RegularLowHand('AcAsQdQhKc')
        >>> h0 == h1
        True
        >>> h0 = RegularLowHand.from_game('AcAd', 'AhAsKcQd')
        >>> h1 = RegularLowHand('AdAhAsKcQd')
        >>> h0 == h1
        True

        :param hole_cards: The hole cards.
        :param board_cards: The optional board cards.
        :return: The strongest hand from possible card combinations.
        """
        max_hand = None

        for combination in combinations(
                chain(Card.clean(hole_cards), Card.clean(board_cards)),
                cls.card_count,
        ):
            try:
                hand = cls(combination)
            except ValueError:
                pass
            else:
                if max_hand is None or hand > max_hand:
                    max_hand = hand

        if max_hand is None:
            raise ValueError('no valid hand')

        return max_hand


class StandardHand(CombinationHand, ABC):
    """The abstract base class for standard hands."""

    lookup = StandardLookup()
    card_count = 5


class StandardHighHand(StandardHand):
    """The class for standard high hands.

    >>> h0 = StandardHighHand('7c5d4h3s2c')
    >>> h1 = StandardHighHand('7c6d4h3s2c')
    >>> h2 = StandardHighHand('8c7d6h4s2c')
    >>> h3 = StandardHighHand('AcAsAd2s4s')
    >>> h4 = StandardHighHand('TsJsQsKsAs')
    >>> h0 < h1 < h2 < h3 < h4
    True

    >>> h = StandardHighHand('4c5dThJsAcKh2h')
    Traceback (most recent call last):
        ...
    ValueError: invalid hand '4c5dThJsAcKh2h'
    >>> h = StandardHighHand('Ac2c3c4c')
    Traceback (most recent call last):
        ...
    ValueError: invalid hand 'Ac2c3c4c'
    >>> h = StandardHighHand(())
    Traceback (most recent call last):
        ...
    ValueError: invalid hand ''
    """

    low = False


class StandardLowHand(StandardHand):
    """The class for standard low hands.

    >>> h0 = StandardLowHand('TsJsQsKsAs')
    >>> h1 = StandardLowHand('AcAsAd2s4s')
    >>> h2 = StandardLowHand('8c7d6h4s2c')
    >>> h3 = StandardLowHand('7c6d4h3s2c')
    >>> h4 = StandardLowHand('7c5d4h3s2c')
    >>> h0 < h1 < h2 < h3 < h4
    True

    >>> h = StandardLowHand('4c5dThJsAcKh2h')
    Traceback (most recent call last):
        ...
    ValueError: invalid hand '4c5dThJsAcKh2h'
    >>> h = StandardLowHand('Ac2c3c4c')
    Traceback (most recent call last):
        ...
    ValueError: invalid hand 'Ac2c3c4c'
    >>> h = StandardLowHand(())
    Traceback (most recent call last):
        ...
    ValueError: invalid hand ''
    """

    low = True


class ShortDeckHoldemHand(CombinationHand):
    """The class for short-deck hold'em hands.

    Here, flushes beat full houses.

    >>> h0 = ShortDeckHoldemHand('6c7d8h9sJc')
    >>> h1 = ShortDeckHoldemHand('7c7d7hTsQc')
    >>> h2 = ShortDeckHoldemHand('As6c7h8h9h')
    >>> h3 = ShortDeckHoldemHand('AsAhKcKhKd')
    >>> h4 = ShortDeckHoldemHand('6s7s8sTsQs')
    >>> h0 < h1 < h2 < h3 < h4
    True

    >>> h = ShortDeckHoldemHand('4c5dThJsAcKh2h')
    Traceback (most recent call last):
        ...
    ValueError: invalid hand '4c5dThJsAcKh2h'
    >>> h = ShortDeckHoldemHand('Ac2c3c4c5c')
    Traceback (most recent call last):
        ...
    ValueError: invalid hand 'Ac2c3c4c5c'
    >>> h = ShortDeckHoldemHand(())
    Traceback (most recent call last):
        ...
    ValueError: invalid hand ''
    """

    lookup = ShortDeckHoldemLookup()
    low = False
    card_count = 5


class EightOrBetterLowHand(CombinationHand):
    """The class for eight or better low hands.

    >>> h0 = EightOrBetterLowHand('8c7d6h4s2c')
    >>> h1 = EightOrBetterLowHand('7c5d4h3s2c')
    >>> h2 = EightOrBetterLowHand('5d4h3s2dAd')
    >>> h0 < h1 < h2
    True

    >>> h = EightOrBetterLowHand('AcAsAd2s4s')
    Traceback (most recent call last):
        ...
    ValueError: invalid hand 'AcAsAd2s4s'
    >>> h = EightOrBetterLowHand('TsJsQsKsAs')
    Traceback (most recent call last):
        ...
    ValueError: invalid hand 'TsJsQsKsAs'
    >>> h = EightOrBetterLowHand('4c5dThJsAcKh2h')
    Traceback (most recent call last):
        ...
    ValueError: invalid hand '4c5dThJsAcKh2h'
    >>> h = EightOrBetterLowHand('Ac2c3c4c')
    Traceback (most recent call last):
        ...
    ValueError: invalid hand 'Ac2c3c4c'
    >>> h = EightOrBetterLowHand(())
    Traceback (most recent call last):
        ...
    ValueError: invalid hand ''
    """

    lookup = EightOrBetterLookup()
    low = True
    card_count = 5


class RegularLowHand(CombinationHand):
    """The class for low regular hands.

    Here, flushes are ignored.

    >>> h0 = RegularLowHand('KhKsKcKdAc')
    >>> h1 = RegularLowHand('2s2c3s3cAh')
    >>> h2 = RegularLowHand('6c4d3h2sAc')
    >>> h3 = RegularLowHand('Ac2c3c4c5c')
    >>> h0 < h1 < h2 < h3
    True

    >>> h = RegularLowHand('4c5dThJsAcKh2h')
    Traceback (most recent call last):
        ...
    ValueError: invalid hand '4c5dThJsAcKh2h'
    >>> h = RegularLowHand(())
    Traceback (most recent call last):
        ...
    ValueError: invalid hand ''
    """

    lookup = RegularLookup()
    low = True
    card_count = 5


class BoardCombinationHand(CombinationHand, ABC):
    """The abstract base class for board-combination hands."""

    board_card_count: ClassVar[int]
    """The number of board cards."""

    @classmethod
    def from_game(
            cls,
            hole_cards: CardsLike,
            board_cards: CardsLike = (),
    ) -> Hand:
        """Create a poker hand from a game setting.

        In a game setting, a player uses private cards from their hole
        and the public cards from the board to make their hand.

        >>> h0 = GreekHoldemHand.from_game('Ac2d', 'QdJdTh2sKs')
        >>> h1 = GreekHoldemHand('2s2dAcKsQd')
        >>> h0 == h1
        True
        >>> h0 = GreekHoldemHand.from_game('AsKs', 'QdJdTh2s2d')
        >>> h1 = GreekHoldemHand('AsKsQdJdTh')
        >>> h0 == h1
        True
        >>> h0 = GreekHoldemHand.from_game('Ac9c', 'AhKhQhJhTh')
        >>> h1 = GreekHoldemHand('AcAhKhQh9c')
        >>> h0 == h1
        True

        :param hole_cards: The hole cards.
        :param board_cards: The optional board cards.
        :return: The strongest hand from possible card combinations.
        """
        hole_cards = Card.clean(hole_cards)
        board_cards = Card.clean(board_cards)
        max_hand = None

        for combination in combinations(board_cards, cls.board_card_count):
            try:
                hand = super().from_game(hole_cards, combination)
            except ValueError:
                pass
            else:
                if max_hand is None or hand > max_hand:
                    max_hand = hand

        if max_hand is None:
            raise ValueError('no valid hand')

        return max_hand


class GreekHoldemHand(BoardCombinationHand):
    """The class for Greek hold'em hands.

    In Greek hold'em, the player must use all of their hole cards to
    make a hand.

    >>> h0 = GreekHoldemHand('7c5d4h3s2c')
    >>> h1 = GreekHoldemHand('7c6d4h3s2c')
    >>> h2 = GreekHoldemHand('8c7d6h4s2c')
    >>> h3 = GreekHoldemHand('AcAsAd2s4s')
    >>> h4 = GreekHoldemHand('TsJsQsKsAs')
    >>> h0 < h1 < h2 < h3 < h4
    True
    """

    lookup = StandardLookup()
    low = False
    card_count = 5
    board_card_count = 3


class HoleBoardCombinationHand(BoardCombinationHand, ABC):
    """The abstract base class for hole-board-combination hands."""

    hole_card_count: ClassVar[int]
    """The number of hole cards."""

    @classmethod
    def from_game(
            cls,
            hole_cards: CardsLike,
            board_cards: CardsLike = (),
    ) -> Hand:
        """Create a poker hand from a game setting.

        In a game setting, a player uses private cards from their hole
        and the public cards from the board to make their hand.

        >>> h0 = OmahaHoldemHand.from_game('6c7c8c9c', '8s9sTc')
        >>> h1 = OmahaHoldemHand('6c7c8s9sTc')
        >>> h0 == h1
        True
        >>> h0 = OmahaHoldemHand.from_game('6c7c8s9s', '8c9cTc')
        >>> h1 = OmahaHoldemHand('6c7c8c9cTc')
        >>> h0 == h1
        True
        >>> h0 = OmahaHoldemHand.from_game('6c7c8c9c', '8s9sTc9hKs')
        >>> h1 = OmahaHoldemHand('8c8s9c9s9h')
        >>> h0 == h1
        True
        >>> h0 = OmahaHoldemHand.from_game('6c7c8sAh', 'As9cTc2sKs')
        >>> h1 = OmahaHoldemHand('AhAsKsTc8s')
        >>> h0 == h1
        True

        >>> h0 = OmahaEightOrBetterLowHand.from_game('As2s3s4s', '2c3c4c5c6c')
        >>> h1 = OmahaEightOrBetterLowHand('Ad2d3d4d5d')
        >>> h0 == h1
        True
        >>> h0 = OmahaEightOrBetterLowHand.from_game('As6s7s8s', '2c3c4c5c6c')
        >>> h1 = OmahaEightOrBetterLowHand('Ad2d3d4d6d')
        >>> h0 == h1
        True

        :param hole_cards: The hole cards.
        :param board_cards: The optional board cards.
        :return: The strongest hand from possible card combinations.
        """
        hole_cards = Card.clean(hole_cards)
        board_cards = Card.clean(board_cards)
        max_hand = None

        for combination in combinations(hole_cards, cls.hole_card_count):
            try:
                hand = super().from_game(combination, board_cards)
            except ValueError:
                pass
            else:
                if max_hand is None or hand > max_hand:
                    max_hand = hand

        if max_hand is None:
            raise ValueError('no valid hand')

        return max_hand


class OmahaHoldemHand(HoleBoardCombinationHand):
    """The class for Omaha hold'em hands.

    In Omaha hold'em, the player must use a fixed number of their hole
    cards to make a hand.

    >>> h0 = OmahaHoldemHand('7c5d4h3s2c')
    >>> h1 = OmahaHoldemHand('7c6d4h3s2c')
    >>> h2 = OmahaHoldemHand('8c7d6h4s2c')
    >>> h3 = OmahaHoldemHand('AcAsAd2s4s')
    >>> h4 = OmahaHoldemHand('TsJsQsKsAs')
    >>> h0 < h1 < h2 < h3 < h4
    True
    """

    lookup = StandardLookup()
    low = False
    card_count = 5
    board_card_count = 3
    hole_card_count = 2


class OmahaEightOrBetterLowHand(HoleBoardCombinationHand):
    """The class for Omaha eight or better low hands.

    >>> h0 = OmahaEightOrBetterLowHand('8c7d6h4s2c')
    >>> h1 = OmahaEightOrBetterLowHand('7c5d4h3s2c')
    >>> h2 = OmahaEightOrBetterLowHand('5d4h3s2dAd')
    >>> h0 < h1 < h2
    True
    """

    lookup = EightOrBetterLookup()
    low = True
    card_count = 5
    board_card_count = 3
    hole_card_count = 2


class BadugiHand(Hand):
    """The class for badugi hands.

    >>> h0 = BadugiHand('Kc')
    >>> h1 = BadugiHand('Ac')
    >>> h2 = BadugiHand('4c8dKh')
    >>> h3 = BadugiHand('Ac2d3h5s')
    >>> h4 = BadugiHand('Ac2d3h4s')
    >>> h0 < h1 < h2 < h3 < h4
    True

    >>> h = BadugiHand('Ac2d3c4s5c')
    Traceback (most recent call last):
        ...
    ValueError: invalid hand 'Ac2d3c4s5c'
    >>> h = BadugiHand('Ac2d3c4s')
    Traceback (most recent call last):
        ...
    ValueError: cards not rainbow
    >>> h = BadugiHand('AcAd3h4s')
    Traceback (most recent call last):
        ...
    ValueError: invalid hand 'AcAd3h4s'
    >>> h = BadugiHand('Ac2c')
    Traceback (most recent call last):
        ...
    ValueError: invalid hand 'Ac2c'
    >>> h = BadugiHand(())
    Traceback (most recent call last):
        ...
    ValueError: invalid hand ''
    """

    lookup = BadugiLookup()
    low = True

    @classmethod
    def from_game(
            cls,
            hole_cards: CardsLike,
            board_cards: CardsLike = (),
    ) -> Hand:
        """Create a poker hand from a game setting.

        In a game setting, a player uses private cards from their hole
        and the public cards from the board to make their hand.

        >>> h0 = BadugiHand.from_game('2s4c5d6h')
        >>> h1 = BadugiHand('2s4c5d6h')
        >>> h0 == h1
        True
        >>> h0 = BadugiHand.from_game('2s3s4d7h')
        >>> h1 = BadugiHand('2s4d7h')
        >>> h0 == h1
        True
        >>> h0 = BadugiHand.from_game('KcKdKhKs')
        >>> h1 = BadugiHand('Ks')
        >>> h0 == h1
        True

        :param hole_cards: The hole cards.
        :param board_cards: The optional board cards.
        :return: The strongest hand from possible card combinations.
        """
        ranks = set()
        suits = set()
        cards = set()

        for card in sorted(
                chain(Card.clean(hole_cards), Card.clean(board_cards)),
                key=lambda card: RankOrder.STANDARD.index(card.rank),
        ):
            if card.rank not in ranks and card.suit not in suits:
                ranks.add(card.rank)
                suits.add(card.suit)
                cards.add(card)

        return cls(cards)

    def __init__(self, cards: CardsLike) -> None:
        super().__init__(cards)

        if not Card.are_rainbow(self.cards):
            raise ValueError('cards not rainbow')


class KuhnPokerHand(Hand):
    """The class for Kuhn poker hands.

    >>> h0 = KuhnPokerHand('Js')
    >>> h1 = KuhnPokerHand('Qs')
    >>> h2 = KuhnPokerHand('Ks')
    >>> h0 < h1 < h2
    True

    >>> h = KuhnPokerHand('As')
    Traceback (most recent call last):
        ...
    ValueError: invalid hand 'As'
    """

    lookup = KuhnPokerLookup()
    low = False

    @classmethod
    def from_game(
            cls,
            hole_cards: CardsLike,
            board_cards: CardsLike = (),
    ) -> Hand:
        """Create a poker hand from a game setting.

        In a game setting, a player uses private cards from their hole
        and the public cards from the board to make their hand.

        >>> h0 = KuhnPokerHand.from_game('Ks')
        >>> h1 = KuhnPokerHand('Ks')
        >>> h0 == h1
        True

        :param hole_cards: The hole cards.
        :param board_cards: The optional board cards.
        :return: The strongest hand from possible card combinations.
        """
        return max(
            map(cls, chain(Card.clean(hole_cards), Card.clean(board_cards))),
        )



""":mod:`pokerkit.lookups` implements classes related to poker hand
lookups.
"""

from abc import ABC
from collections.abc import Iterable, Reversible, Sequence
from collections import Counter
from dataclasses import dataclass, field
from enum import StrEnum, unique
from functools import partial
from itertools import combinations, filterfalse
from math import prod
from operator import contains
from typing import ClassVar

from pokerkit.utilities import Card, CardsLike, Rank, RankOrder


@unique
class Label(StrEnum):
    """The enum class for all hand classification labels.

    >>> Label.ONE_PAIR
    <Label.ONE_PAIR: 'One pair'>
    >>> Label.FULL_HOUSE
    <Label.FULL_HOUSE: 'Full house'>
    """

    HIGH_CARD: str = 'High card'
    """The label of high cards."""
    ONE_PAIR: str = 'One pair'
    """The label of one pair."""
    TWO_PAIR: str = 'Two pair'
    """The label of two pair."""
    THREE_OF_A_KIND: str = 'Three of a kind'
    """The label of three of a kind."""
    STRAIGHT: str = 'Straight'
    """The label of straight."""
    FLUSH: str = 'Flush'
    """The label of flush."""
    FULL_HOUSE: str = 'Full house'
    """The label of full house."""
    FOUR_OF_A_KIND: str = 'Four of a kind'
    """The label of four of a kind."""
    STRAIGHT_FLUSH: str = 'Straight flush'
    """The label of straight flush."""


@dataclass(order=True, frozen=True)
class Entry:
    """The class for hand lookup entries.

    The :attr:`index` represents the strength of the corresponding
    hand. In this library, a stronger hand is considered greater. In
    other words, stronger hands have a greater index with which
    different entries and hands are compared.

    The attributes are read-only.

    Note that the entries in the example below are meaningless. They
    are only meant to show how entries are used by other poker
    utilities.

    >>> e0 = Entry(0, Label.HIGH_CARD)
    >>> e1 = Entry(1, Label.HIGH_CARD)
    >>> e2 = Entry(1, Label.HIGH_CARD)
    >>> e0 < e1
    True
    >>> e1 < e2
    False
    >>> e0 == e1
    False
    >>> e1 == e2
    True
    >>> e0.index
    0
    >>> e0.label
    <Label.HIGH_CARD: 'High card'>
    """

    index: int
    """The index of the corresponding hand."""
    label: Label = field(compare=False)
    """The label of the corresponding hand."""


@dataclass
class Lookup(ABC):
    """The abstract base class for hand lookups.

    Lookups are used internally by hands. If you want to evaluate poker
    hands, please use one of the hand classes.

    >>> lookup = StandardLookup()
    >>> e0 = lookup.get_entry('As3sQhJsJc')
    >>> e1 = lookup.get_entry('2s4sKhKsKc')
    >>> e0 < e1
    True
    >>> e0.label
    <Label.ONE_PAIR: 'One pair'>
    >>> e1.label
    <Label.THREE_OF_A_KIND: 'Three of a kind'>
    """

    __primes = 2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41

    assert len(__primes) >= len(tuple(Rank)) - 1  # except unknown

    __multipliers = dict(zip(Rank, __primes))
    rank_order: ClassVar[RankOrder]
    """The rank order."""
    __entries: dict[tuple[int, bool], Entry] = field(
        default_factory=dict,
        init=False,
        repr=False,
    )
    __entry_count: int = field(default=0, init=False, repr=False)

    @classmethod
    def __hash(cls, ranks: Iterable[Rank]) -> int:
        return prod(map(cls.__multipliers.__getitem__, ranks))

    @classmethod
    def __hash_multisets(
            cls,
            ranks: Reversible[Rank],
            counter: Counter[int],
    ) -> Sequence[int]:
        if not counter:
            return (cls.__hash(()),)

        hashes = []
        multiplicity = max(counter)
        count = counter.pop(multiplicity)

        for samples in combinations(reversed(ranks), count):
            hash_ = cls.__hash(samples) ** multiplicity

            for partial_hash in cls.__hash_multisets(
                    tuple(filterfalse(partial(contains, samples), ranks)),
                    counter,
            ):
                hashes.append(hash_ * partial_hash)

        counter[multiplicity] = count

        return hashes

    def has_entry(self, cards: CardsLike) -> bool:
        """Return whether the cards can be looked up.

        The cards can be looked up if the lookup contains an entry with
        the given cards.

        >>> lookup = ShortDeckHoldemLookup()
        >>> lookup.has_entry('Ah6h7s8c9s')
        True
        >>> lookup.has_entry('Ah6h7s8c2s')
        False

        :param cards: The cards to look up.
        :return: ``True`` if the cards can looked up, otherwise
                 ``False``.
        """
        return self.__get_key(cards) in self.__entries

    def get_entry(self, cards: CardsLike) -> Entry:
        """Return the corresponding lookup entry of the hand that the
        cards form.

        >>> lookup = ShortDeckHoldemLookup()
        >>> entry = lookup.get_entry('Ah6h7s8c9s')
        >>> entry.index
        1134
        >>> entry.label
        <Label.STRAIGHT: 'Straight'>
        >>> entry = lookup.get_entry('Ah6h7s8c2s')
        Traceback (most recent call last):
            ...
        ValueError: cards form an invalid hand

        :param cards: The cards to look up.
        :return: The corresponding lookup entry.
        :raises ValueError: If cards do not form a valid hand.
        """
        key = self.__get_key(cards)

        if key not in self.__entries:
            raise ValueError('cards form an invalid hand')

        return self.__entries[key]

    def get_entry_or_none(self, cards: CardsLike) -> Entry | None:
        """Return the corresponding lookup entry of the hand that the
        cards form if it exists. Otherwise, return ``None``.

        >>> lookup = ShortDeckHoldemLookup()
        >>> lookup.get_entry('Ah6h7s8c2s')
        Traceback (most recent call last):
            ...
        ValueError: cards form an invalid hand
        >>> lookup.get_entry_or_none('Ah6h7s8c2s') is None
        True

        :param cards: The cards to look up.
        :return: The optional corresponding lookup entry.
        """
        return self.__entries.get(self.__get_key(cards))

    def __get_key(self, cards: CardsLike) -> tuple[int, bool]:
        cards = Card.clean(cards)
        hash_ = self.__hash(Card.get_ranks(cards))
        suitedness = Card.are_suited(cards)

        return hash_, suitedness

    def _add_multisets(
            self,
            counter: Counter[int],
            suitednesses: tuple[bool, ...],
            label: Label,
    ) -> None:
        hashes = self.__hash_multisets(self.rank_order, counter)

        for hash_ in reversed(hashes):
            self.__add(hash_, suitednesses, label)

    def _add_straights(
            self,
            count: int,
            suitednesses: tuple[bool, ...],
            label: Label,
    ) -> None:
        self.__add(
            self.__hash(self.rank_order[-1:] + self.rank_order[: count - 1]),
            suitednesses,
            label,
        )

        for i in range(len(self.rank_order) - count + 1):
            self.__add(
                self.__hash(self.rank_order[i:i + count]),
                suitednesses,
                label,
            )

    def __add(
            self,
            hash_: int,
            suitednesses: Iterable[bool],
            label: Label,
    ) -> None:
        entry = Entry(self.__entry_count, label)
        self.__entry_count += 1

        for suitedness in suitednesses:
            self.__entries[hash_, suitedness] = entry


@dataclass
class StandardLookup(Lookup):
    """The class for standard hand lookups.

    Lookups are used by evaluators. If you want to evaluate poker hands,
    please subclasses of :class:`pokerkit.hands.Hand` that use this
    lookup.

    >>> lookup = StandardLookup()
    >>> e0 = lookup.get_entry('Ah6h7s8c9s')
    >>> e1 = lookup.get_entry('AhAc6s6hTd')
    >>> e2 = lookup.get_entry('AcAdAhAsAc')
    Traceback (most recent call last):
        ...
    ValueError: cards form an invalid hand
    >>> e0 < e1
    True
    >>> e0.label
    <Label.HIGH_CARD: 'High card'>
    >>> e1.label
    <Label.TWO_PAIR: 'Two pair'>
    """

    rank_order = RankOrder.STANDARD

    def __post_init__(self) -> None:
        self._add_multisets(Counter({1: 5}), (False,), Label.HIGH_CARD)
        self._add_multisets(Counter({2: 1, 1: 3}), (False,), Label.ONE_PAIR)
        self._add_multisets(Counter({2: 2, 1: 1}), (False,), Label.TWO_PAIR)
        self._add_multisets(
            Counter({3: 1, 1: 2}),
            (False,),
            Label.THREE_OF_A_KIND,
        )
        self._add_straights(5, (False,), Label.STRAIGHT)
        self._add_multisets(Counter({1: 5}), (True,), Label.FLUSH)
        self._add_multisets(Counter({3: 1, 2: 1}), (False,), Label.FULL_HOUSE)
        self._add_multisets(
            Counter({4: 1, 1: 1}),
            (False,),
            Label.FOUR_OF_A_KIND,
        )
        self._add_straights(5, (True,), Label.STRAIGHT_FLUSH)


@dataclass
class ShortDeckHoldemLookup(Lookup):
    """The class for short-deck hold'em hand lookups.

    Here, flushes beat full houses.

    Lookups are used by evaluators. If you want to evaluate poker hands,
    please use :class:`pokerkit.hands.ShortDeckHoldemHand`.

    >>> lookup = ShortDeckHoldemLookup()
    >>> e0 = lookup.get_entry('AhAc6s6hTd')
    >>> e1 = lookup.get_entry('Ah6h7s8c9s')
    >>> e2 = lookup.get_entry('Ah2h3s4c5s')
    Traceback (most recent call last):
        ...
    ValueError: cards form an invalid hand
    >>> e0 < e1
    True
    >>> e0.label
    <Label.TWO_PAIR: 'Two pair'>
    >>> e1.label
    <Label.STRAIGHT: 'Straight'>
    """

    rank_order = RankOrder.SHORT_DECK_HOLDEM

    def __post_init__(self) -> None:
        self._add_multisets(Counter({1: 5}), (False,), Label.HIGH_CARD)
        self._add_multisets(Counter({2: 1, 1: 3}), (False,), Label.ONE_PAIR)
        self._add_multisets(Counter({2: 2, 1: 1}), (False,), Label.TWO_PAIR)
        self._add_multisets(
            Counter({3: 1, 1: 2}),
            (False,),
            Label.THREE_OF_A_KIND,
        )
        self._add_straights(5, (False,), Label.STRAIGHT)
        self._add_multisets(Counter({3: 1, 2: 1}), (False,), Label.FULL_HOUSE)
        self._add_multisets(Counter({1: 5}), (True,), Label.FLUSH)
        self._add_multisets(
            Counter({4: 1, 1: 1}),
            (False,),
            Label.FOUR_OF_A_KIND,
        )
        self._add_straights(5, (True,), Label.STRAIGHT_FLUSH)


@dataclass
class EightOrBetterLookup(Lookup):
    """The class for eight or better hand lookups.

    Lookups are used by evaluators. If you want to evaluate poker hands,
    please use :class:`pokerkit.hands.EightOrBetterLowHand`.
    """

    rank_order = RankOrder.EIGHT_OR_BETTER_LOW

    def __post_init__(self) -> None:
        self._add_multisets(Counter({1: 5}), (False, True), Label.HIGH_CARD)


@dataclass
class RegularLookup(Lookup):
    """The class for regular hand lookups.

    Here, flushes are ignored.

    Lookups are used by evaluators. If you want to evaluate poker hands,
    please use :class:`pokerkit.hands.RegularLowHand`.

    >>> lookup = RegularLookup()
    >>> e0 = lookup.get_entry('Ah6h7s8c9s')
    >>> e1 = lookup.get_entry('AhAc6s6hTd')
    >>> e2 = lookup.get_entry('3s4sQhTc')
    Traceback (most recent call last):
        ...
    ValueError: cards form an invalid hand
    >>> e0 < e1
    True
    >>> e0.label
    <Label.HIGH_CARD: 'High card'>
    >>> e1.label
    <Label.TWO_PAIR: 'Two pair'>
    """

    rank_order = RankOrder.REGULAR

    def __post_init__(self) -> None:
        self._add_multisets(Counter({1: 5}), (False, True), Label.HIGH_CARD)
        self._add_multisets(Counter({2: 1, 1: 3}), (False,), Label.ONE_PAIR)
        self._add_multisets(Counter({2: 2, 1: 1}), (False,), Label.TWO_PAIR)
        self._add_multisets(
            Counter({3: 1, 1: 2}),
            (False,),
            Label.THREE_OF_A_KIND,
        )
        self._add_multisets(Counter({3: 1, 2: 1}), (False,), Label.FULL_HOUSE)
        self._add_multisets(
            Counter({4: 1, 1: 1}),
            (False,),
            Label.FOUR_OF_A_KIND,
        )


@dataclass
class BadugiLookup(Lookup):
    """The class for badugi hand lookups.

    Lookups are used by evaluators. If you want to evaluate poker hands,
    please use :class:`pokerkit.hands.BadugiHand`.

    >>> lookup = BadugiLookup()
    >>> e0 = lookup.get_entry('2s')
    >>> e1 = lookup.get_entry('KhQc')
    >>> e2 = lookup.get_entry('AcAdAhAs')
    Traceback (most recent call last):
        ...
    ValueError: cards form an invalid hand
    >>> e0 > e1
    True
    >>> e0.label
    <Label.HIGH_CARD: 'High card'>
    >>> e1.label
    <Label.HIGH_CARD: 'High card'>
    """

    rank_order = RankOrder.REGULAR

    def __post_init__(self) -> None:
        for i in range(4, 0, -1):
            self._add_multisets(Counter({1: i}), (i == 1,), Label.HIGH_CARD)


@dataclass
class KuhnPokerLookup(Lookup):
    """The class for Kuhn poker hand lookups.

    Lookups are used by evaluators. If you want to evaluate poker hands,
    please use :class:`pokerkit.hands.KuhnPokerHand`.

    >>> lookup = KuhnPokerLookup()
    >>> e0 = lookup.get_entry('J?')
    >>> e1 = lookup.get_entry('Q?')
    >>> e2 = lookup.get_entry('2?')
    Traceback (most recent call last):
        ...
    ValueError: cards form an invalid hand
    >>> e0 < e1
    True
    >>> e0.label
    <Label.HIGH_CARD: 'High card'>
    >>> e1.label
    <Label.HIGH_CARD: 'High card'>
    """

    rank_order = RankOrder.KUHN_POKER

    def __post_init__(self) -> None:
        self._add_multisets(Counter({1: 1}), (True,), Label.HIGH_CARD)
